﻿using System;
using System.Collections.Generic;
using System.Linq;
using Clarion;
using Clarion.Framework;
using Clarion.Framework.Core;
using Clarion.Framework.Templates;
using WorldServerLibrary.Model;
using System.Threading;

namespace ClarionSimulation
{
    /// <summary>
    /// This event will be fired at the beginning of each cognitive cycle in order to get the current visual sensor information.
    /// </summary>
    /// <returns>The Visual Sensor Information Itens</returns> 
    public delegate IList<Thing> InputVisualSensorialInformationEventHandler();

    /// <summary>
    /// This event will be fired at the end of each cognitive cycle in order to translate the selected action to a World Server Command
    /// </summary>
    /// <param name="externalAction">The external action that was choice</param>
    public delegate void OutputActionChunkEventHandler(ClarionAgentActionType externalAction);

    /// <summary>
    /// Public enum that represents all possibilities of agent actions
    /// </summary>
    public enum ClarionAgentActionType
    {
        DO_NOTHING,
		GO_AHEAD,
        ROTATE_CLOCKWISE,
		EXPLORE,
		GOTOJEWEL,
		GET_JEWEL
    }

    public class ClarionAgent
    {
        #region Constants

        /// <summary>
        /// Constant that represents the Visual Sensor
        /// </summary>
        private String SENSOR_VISUAL_DIMENSION = "VisualSensor";

        /// <summary>
        /// Constant that represents that there is at least one wall ahead
        /// </summary>
        
		private String DIMENSION_WALL_AHEAD = "WallAhead";
		private String DIMENSION_GOTOJEWEL = "GoToJewel";
		private String DIMENSION_JEWEL_AHEAD = "JewellAhead";
		private String DIMENSION_EXPLORE = "Explore";


        #endregion

        #region Properties

        #region Simulation

        /// <summary>
        /// If this value is greater than zero, the agent will have a finite number of cognitive cycle. Otherwise, it will have infinite cycles.
        /// </summary>
        public double MaxNumberOfCognitiveCycles = -1;

        /// <summary>
        /// Current cognitive cycle number
        /// </summary>
        private double CurrentCognitiveCycle = 0;

        /// <summary>
        /// Time between cognitive cycle in miliseconds
        /// </summary>
        public Int32 TimeBetweenCognitiveCycles = 0;

        /// <summary>
        /// A thread Class that will handle the simulation process
        /// </summary>
        private Thread runThread;

        #endregion

        #region Agent

        /// <summary>
        /// The agent 
        /// </summary>
        private Clarion.Framework.Agent CurrentAgent;

        /// <summary>
        /// For each cognitive cycle, this event will be called in order to the agent receives the current sensorial information
        /// </summary>
        public event InputVisualSensorialInformationEventHandler OnNewVisualSensorialInformation;

        /// <summary>
        /// For each cognitive cycle, this event will be called when the agent selects one action
        /// </summary>
        public event OutputActionChunkEventHandler OnNewExternalActionSelected;

        #endregion

        #region Perception Input

        /// <summary>
        /// Perception input to indicates a wall ahead
        /// </summary>
        
		private DimensionValuePair InputWallAhead;

		private DimensionValuePair InputGoToJewel; 

		//# Declarar o input de entrada das joias
		private DimensionValuePair InputGetJewel;

		// #Declarar o input para ficar rotacionando
		private DimensionValuePair InputExplore;
        #endregion

        #region Action Output

        /// <summary>
        /// Output action that makes the agent to rotate clockwise
        /// </summary>
		private ExternalActionChunk OutputRotateClockwise;
        
		/// <summary>
        /// Output action that makes the agent go ahead
        /// </summary>
        private ExternalActionChunk OutputGoAhead;

		//# Declarar a ação externa para pegar a joia
		private ExternalActionChunk OutputGetJewel;

		//# Ação de explorar o ambiente
		private ExternalActionChunk OutputExplore;

		//#Ação para andar até a joia
		private ExternalActionChunk OutputGoToJewel;


        #endregion

        #endregion

        #region Constructor

        public ClarionAgent()
        {

            // Initialize the agent
            CurrentAgent = World.NewAgent("Current Agent");

            // Initialize Input Information
            InputWallAhead = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_WALL_AHEAD);

			// #colocar a informação da joia no agente
			InputGetJewel = World.NewDimensionValuePair(SENSOR_VISUAL_DIMENSION, DIMENSION_JEWEL_AHEAD); 

			//#coloca informação para explorar o ambiente
			InputExplore = World.NewDimensionValuePair (SENSOR_VISUAL_DIMENSION, DIMENSION_EXPLORE);

			InputGoToJewel = World.NewDimensionValuePair (SENSOR_VISUAL_DIMENSION, DIMENSION_GOTOJEWEL);

            // Initialize Output actions
            OutputRotateClockwise = World.NewExternalActionChunk(ClarionAgentActionType.ROTATE_CLOCKWISE.ToString());
            
			OutputGoAhead = World.NewExternalActionChunk(ClarionAgentActionType.GO_AHEAD.ToString());

			//#colocar a informação para o agente ir até a joia
			OutputGoToJewel=World.NewExternalActionChunk(ClarionAgentActionType.GOTOJEWEL.ToString());

			// #colocar a informação para o agente pegar a joia
			OutputGetJewel = World.NewExternalActionChunk (ClarionAgentActionType.GET_JEWEL.ToString());

			//#explorar o ambiente
			OutputExplore = World.NewExternalActionChunk(ClarionAgentActionType.EXPLORE.ToString());

            //Create thread to simulation
            runThread = new Thread(RunThread);
			Console.WriteLine("Agent started");
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Run the Simulation in World Server 3d Environment
        /// </summary>
        public void Run()
        {                
			Console.WriteLine ("Running ...");
            // Setup Agent to run
            if (runThread != null && !runThread.IsAlive)
            {
                SetupAgentInfraStructure();

                if (OnNewVisualSensorialInformation != null)
                {
                    // Start Simulation Thread                
                    runThread.Start(null);
                }
            }
        }

        /// <summary>
        /// Abort the current Simulation
        /// </summary>
        /// <param name="deleteAgent">If true beyond abort the current simulation it will die the agent.</param>
        public void Abort(Boolean deleteAgent)
        {   Console.WriteLine ("Aborting ...");
            if (runThread != null && runThread.IsAlive)
            {
                runThread.Abort();
            }

            if (CurrentAgent != null && deleteAgent)
            {
                CurrentAgent.Die();
            }
        }

        #endregion

        #region Setup Agent Methods

        /// <summary>
        /// Setup agent infra structure (ACS, NACS, MS and MCS)
        /// </summary>
        private void SetupAgentInfraStructure()
        {
            // Setup the ACS Subsystem
            SetupACS();                    
        }

        private void SetupMS()
        {            
            //RichDrive
        }

        /// <summary>
        /// Setup the ACS subsystem
        /// </summary>
        private void SetupACS()
        {
     
			// Create Colission Wall Rule
            SupportCalculator avoidCollisionWallSupportCalculator = FixedRuleDelegateToAvoidColissionWall;
            FixedRule ruleAvoidCollisionWall = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputRotateClockwise, avoidCollisionWallSupportCalculator);
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleAvoidCollisionWall);
            // Create Colission To Go Ahead
            SupportCalculator goAheadSupportCalculator = FixedRuleDelegateToGoAhead;
            FixedRule ruleGoAhead = AgentInitializer.InitializeActionRule(CurrentAgent, FixedRule.Factory, OutputGoAhead, goAheadSupportCalculator);          
            // Commit this rule to Agent (in the ACS)
            CurrentAgent.Commit(ruleGoAhead);

			// #Criar uma regra para pegar as joias
			SupportCalculator getJewelSupportCalculator = FixedRuleDelegateToGetJewel;
			FixedRule ruleGetJewel = AgentInitializer.InitializeActionRule (CurrentAgent, FixedRule.Factory, OutputGetJewel, getJewelSupportCalculator);
			// Commit this rule to Agent (in the ACS)
			CurrentAgent.Commit(ruleGetJewel);

			//# Regra para explorar o ambiente em busca de uma jóia
			SupportCalculator exploreSupportCalculator = FixedRuleDelegateToExplore;
			FixedRule ruleExplore = AgentInitializer.InitializeActionRule (CurrentAgent, FixedRule.Factory, OutputExplore, exploreSupportCalculator);
			// Commit this rule to Agent (in the ACS)
			CurrentAgent.Commit(ruleExplore);

			//#Regra para ir até a joia
			SupportCalculator goToJewelSupportCalculator = FixedRuleDelegateGoToJewel;
			FixedRule ruleGoToJewel = AgentInitializer.InitializeActionRule (CurrentAgent, FixedRule.Factory, OutputGoToJewel, goToJewelSupportCalculator);
			// Commit this rule to Agent (in the ACS)
			CurrentAgent.Commit(ruleGoToJewel);


            // Disable Rule Refinement
            CurrentAgent.ACS.Parameters.PERFORM_RER_REFINEMENT = false;

            // The selectio type will be probabilistic
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_METHOD = ActionCenteredSubsystem.LevelSelectionMethods.STOCHASTIC;

            // The selection of the action will be fixed (not variable) i.e. only the statement defined above.
            CurrentAgent.ACS.Parameters.LEVEL_SELECTION_OPTION = ActionCenteredSubsystem.LevelSelectionOptions.FIXED;

            // Define Probabilistic values
            CurrentAgent.ACS.Parameters.FIXED_FR_LEVEL_SELECTION_MEASURE = 1;
            CurrentAgent.ACS.Parameters.FIXED_IRL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_BL_LEVEL_SELECTION_MEASURE = 0;
            CurrentAgent.ACS.Parameters.FIXED_RER_LEVEL_SELECTION_MEASURE = 0;
        }

        /// <summary>
        /// Make the agent perception. In other words, translate the information that came from sensors to a new type that the agent can understand
        /// </summary>
        /// <param name="sensorialInformation">The information that came from server</param>
        /// <returns>The perceived information</returns>
        private SensoryInformation MakePerceptionFromSensorialInput(IList<Thing> sensorialInformation)
		{
			// New sensory information
			SensoryInformation si = World.NewSensoryInformation (CurrentAgent);

			Creature creature = null;
			foreach (Thing item in sensorialInformation) {
				if (item.CategoryId == Thing.CATEGORY_CREATURE) {
					creature = (Creature)item;
				}
			}

			if (sensorialInformation.Where (item => ((item.CategoryId == Thing.CATEGORY_JEWEL) && item.DistanceToCreature <= 28.0)).Any ()) {
				si.Add (InputWallAhead, 0.0);
				si.Add (InputGetJewel, 1.0);
				si.Add (InputExplore, 0.0);
				si.Add (InputGoToJewel, 0.0);
			} else {
				if (sensorialInformation.Where (item => ((item.CategoryId == Thing.CATEGORY_JEWEL) && item.DistanceToCreature > 28.0)).Any ()) {
					si.Add (InputWallAhead, 0.0);
					si.Add (InputGetJewel, 0.0);
					si.Add (InputExplore, 0.0);
					si.Add (InputGoToJewel, 1.0);
				} else {
					if (sensorialInformation.Where (item => (item.CategoryId == Thing.CATEGORY_BRICK && item.DistanceToCreature <= 61)).Any ()) {
						si.Add (InputWallAhead, 1.0);
						si.Add (InputGoToJewel, 0.0);
						si.Add (InputGetJewel, 0.0);
						si.Add (InputExplore, 0.0);
					} else {
						if (sensorialInformation.Where (item => (item.CategoryId == Thing.CATEGORY_BRICK && item.DistanceToCreature > 61)).Any ()) {
							si.Add (InputWallAhead, 1.0);
							si.Add (InputGoToJewel, 0.0);
							si.Add (InputGetJewel, 0.0);
							si.Add (InputExplore, 0.0);
						}
					}
				}
			}
				

			Creature c = (Creature) sensorialInformation.Where(item => (item.CategoryId == Thing.CATEGORY_CREATURE)).First();
			c.PrintLeaflet();
			return si;
        }

        #endregion

        #region Delegate Methods

        #region Fixed Rules

        private double FixedRuleDelegateToAvoidColissionWall(ActivationCollection currentInput, Rule target)
        {
            // See partial match threshold to verify what are the rules available for action selection
            return ((currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
        }

        private double FixedRuleDelegateToGoAhead(ActivationCollection currentInput, Rule target)
        {
            // Here we will make the logic to go ahead
            return ((currentInput.Contains(InputWallAhead, CurrentAgent.Parameters.MIN_ACTIVATION))) ? 1.0 : 0.0;
        }

		private double FixedRuleDelegateToGetJewel(ActivationCollection currentInput, Rule target)
		{
			// Colocar o logica para pegar a joia aqui
			if((currentInput.Contains(InputGetJewel, CurrentAgent.Parameters.MAX_ACTIVATION))){
				return 1.0;
			}else{
				return 0.0;
			}
		}

		private double FixedRuleDelegateGoToJewel(ActivationCollection currentInput, Rule target)
		{
			return ((currentInput.Contains(InputGoToJewel, CurrentAgent.Parameters.MAX_ACTIVATION))) ? 1.0 : 0.0;
		}

		private double FixedRuleDelegateToExplore(ActivationCollection currentInput, Rule target)
		{
			if(currentInput.Contains(InputGetJewel, CurrentAgent.Parameters.MIN_ACTIVATION)	
					&& (currentInput.Contains(InputGoToJewel, CurrentAgent.Parameters.MIN_ACTIVATION)))			   
				{
					return 1.0;
				}else{ 
					return 0.0;
				}
		}
				

        #endregion

        #region Run Thread Method

        private void RunThread(object obj)
        {
			Console.WriteLine("Starting Cognitive Cycle ... press CTRL-C to finish !");
            // Cognitive Cycle starts here getting sensorial information
            while (CurrentCognitiveCycle != MaxNumberOfCognitiveCycles)
            {   
                // Get current sensorial information                    
                IList<Thing> sensorialInformation = OnNewVisualSensorialInformation();

                // Make the perception
                SensoryInformation si = MakePerceptionFromSensorialInput(sensorialInformation);

                //Perceive the sensory information
                CurrentAgent.Perceive(si);

                //Choose an action
                ExternalActionChunk chosen = CurrentAgent.GetChosenExternalAction(si);

                // Get the selected action
                 String actionLabel = chosen.LabelAsIComparable.ToString();
                 ClarionAgentActionType actionType = (ClarionAgentActionType)Enum.Parse(typeof(ClarionAgentActionType), actionLabel, true);

                // Call the output event handler
                if (OnNewExternalActionSelected != null)
                {
                    OnNewExternalActionSelected(actionType);
                }

                // Increment the number of cognitive cycles
                CurrentCognitiveCycle++;

                //Wait to the agent accomplish his job
                if (TimeBetweenCognitiveCycles > 0)
                {
                    Thread.Sleep(TimeBetweenCognitiveCycles);
                }
            }
        }

        #endregion

        #endregion
    }
}
